# TODO this should not be needed in the future
name_to_id_map = {
    'NOWPreprocessor': 'vlxgoxlv',
    'NOWQdrantIndexer16': '11yxatzh',
    'NOWElasticIndexer': 'mdtuyaaj',
    'NOWAutoCompleteExecutor2': 'w5w084h7',
}

id_to_name_map = {
    'vlxgoxlv': 'NOWPreprocessor',
    '11yxatzh': 'NOWQdrantIndexer16',
    'mdtuyaaj': 'NOWElasticIndexer',
    'w5w084h7': 'NOWAutoCompleteExecutor2',
}
